/* $Header: d:/cvsroot/tads/html/htmldbg.h,v 1.1.1.1 1999/05/09 21:32:27 MJRoberts Exp $ */

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  htmldbg.h - HTML TADS debugger interface
Function
  This class implements a set of portable operations intended to simplify
  implementation of the HTML TADS debugger on each platform.  Most of
  the debugger user interface is port-specific code; this class contains
  portable helper routines that the user interface code will usually
  need.

  This class is meant to be instantiated once, so it's probably easiest
  for the the main application frame or the main debugger window to
  create and own an instance of this object.
Notes
  
Modified
  02/04/98 MJRoberts  - Creation
*/

#ifndef HTMLDBG_H
#define HTMLDBG_H

#ifndef TADSHTML_H
#include "tadshtml.h"
#endif

#include <os.h>
#include <dbg.h>


/* size of buffer we need for building a breakpoint location string */
const int HTMLDBG_MAXBPNAME = OSFNMAX + 30;

/* maximum length of a condition string */
const int HTMLDBG_MAXCOND = 256;

/* maximum length of a breakpoint description */
const int HTMLDBG_MAXBPDESC = 512;

/* ------------------------------------------------------------------------ */
/*
 *   System services interfaces for debugger helper functions.  Certain
 *   debugger helper functions require certain operations to be performed
 *   by the user interface.  When calling such functions, the debugger's
 *   system-specific code must provide implementations of some of these
 *   interfaces.  
 */

/* window type specifier */
enum HtmlDbg_wintype_t
{
    HTMLDBG_WINTYPE_SRCFILE,                          /* source file window */
    HTMLDBG_WINTYPE_STACK,                                  /* stack window */
    HTMLDBG_WINTYPE_HISTORY,                   /* call history trace window */
    HTMLDBG_WINTYPE_DEBUGLOG                            /* debug log window */
};

/*
 *   Window creation interface.  The debugger system-specific code must
 *   provide an implementation of this window creation interface when
 *   calling helper routines that need to open new debugger windows.  
 */
class CHtmlDebugSysIfc_win
{
public:
    /*
     *   Create a new system window using the given parser and formatter
     *   instances.  'win_type' specifies the type of the window to
     *   create; if the system code wants to use different subclasses for
     *   the different types, it can use this type specifier to determine
     *   the type to create.  
     */
    virtual class CHtmlSysWin
        *dbgsys_create_win(class CHtmlParser *parser,
                           class CHtmlFormatter *formatter,
                           const textchar_t *win_title,
                           HtmlDbg_wintype_t win_type) = 0;

    /*
     *   Adjust a font descriptor to select the appropriate font for a
     *   source window.  The font descriptor will be set up with for a
     *   default fixed-pitch font; if the system code wants to make any
     *   further adjustments, such as setting a new font size, it can do
     *   so here.  
     */
    virtual void dbgsys_set_srcfont(class CHtmlFontDesc *font_desc) = 0;

    /*
     *   Compare two filenames to see if they actually refer to the same
     *   file.  This should consider system-specific factors such as case
     *   sensitivity, and should also consider relative paths.  If the two
     *   filenames refer to the same file after resolving relative paths
     *   and other factors, this should return true, otherwise false. 
     */
    virtual int dbgsys_cmp_fnames(const char *fname1, const char *fname2) = 0;
};

/* ------------------------------------------------------------------------ */
/*
 *   Debugger helper class
 */

class CHtmlDebugHelper
{
public:
    CHtmlDebugHelper();
    ~CHtmlDebugHelper();

    /*
     *   Update the debugger for a change in the execution status.  This
     *   should be called whenever we re-enter the debugger after
     *   executing game code.  We'll show the current line, reset the
     *   evaluation context, and update the stack window.  Returns the
     *   activated window.  
     */
    class CHtmlSysWin
        *update_for_entry(struct dbgcxdef *ctx,
                          class CHtmlDebugSysIfc_win *syswinifc);

    /*
     *   Set the debugging execution state.  These calls adjust the TADS
     *   interpreter state to the selected debugging state.  Since these
     *   state changes are usually made from within the user interface
     *   while the debugger has control, the debugger UI should normally
     *   exit its event loop (which will return control to the interpreter
     *   and thus to the game being executed) after making one of these
     *   calls. 
     */

    /* set execution state to GO */
    void set_exec_state_go(struct dbgcxdef *ctx);

    /* set execution state to STEP OVER - steps without following calls */
    void set_exec_state_step_over(struct dbgcxdef *ctx);

    /* set execution state to STEP INTO - steps and follows calls */
    void set_exec_state_step_into(struct dbgcxdef *ctx);

    /* 
     *   set execution state to STEP OUT - steps until current function
     *   has returned 
     */
    void set_exec_state_step_out(struct dbgcxdef *ctx);

    /*
     *   Move execution to the current selection in the given window.
     *   Returns zero on success, non-zero on error.  If this is
     *   successful, the debugger should immediately execute a single
     *   step, which will simply break back into the debugger immediately
     *   at this line; the single-step execution will ensure that we set
     *   up the debugger frame correctly at the new statement.  
     */
    int set_next_statement(struct dbgcxdef *ctx, unsigned int *exec_ofs,
                           class CHtmlSysWin *win);

    /*
     *   Check if the current selection in the given window is within the
     *   same function or method as the current execution point.  Returns
     *   true if so, false if not.  This is a condition of setting the
     *   next statement; this function is provided so that the UI code can
     *   check to make sure that this condition is met when the user
     *   attempts to move the execution point, and generate an appropriate
     *   error message if not.  
     */
    int is_in_same_fn(struct dbgcxdef *ctx, unsigned int *exec_ofs,
                      class CHtmlSysWin *win);

    /*
     *   Signal a QUIT condition.  Because this uses the TADS error
     *   signalling mechanism, the caller should take care only to call
     *   this from the point at which its own event loop would normally be
     *   about to return, because this will invoke longjmp() to jump out
     *   of the caller's stack context. 
     */
    static void signal_quit(struct dbgcxdef *ctx);

    /*
     *   Signal a RESTART condition, restarting the game from the
     *   beginning.  The same restrictions that apply to signal_quit()
     *   apply here.  
     */
    static void signal_restart(struct dbgcxdef *ctx);

    /*
     *   Signal an ABORT condition.  This will abort the current command,
     *   returning control to the game's command line for a new command.
     *   The same restrictions that apply to signal_quit() apply here.  
     */
    static void signal_abort(struct dbgcxdef *ctx);

    /*
     *   Get/set the call tracing status 
     */
    int is_call_trace_active(struct dbgcxdef *ctx) const;
    void set_call_trace_active(struct dbgcxdef *ctx, int flag);

    /*
     *   Clear the call trace history log 
     */
    void clear_call_trace_log(struct dbgcxdef *ctx);

    /*
     *   Get a pointer to the history buffer and the number of bytes of
     *   data currently in the log 
     */
    const textchar_t *get_call_trace_buf(struct dbgcxdef *ctx) const;
    unsigned long get_call_trace_len(struct dbgcxdef *ctx) const;

    /*
     *   Set or clear a breakpoint at the current selection in the given
     *   window 
     */
    void toggle_breakpoint(struct dbgcxdef *ctx, class CHtmlSysWin *win);

    /*
     *   Set a temporary breakpoint at the current selection in the given
     *   window.  If there's already a breakpoint at this location, we
     *   won't add a new breakpoint, and we'll return false.  Otherwise,
     *   we'll set a new breakpoint and return true.
     *   
     *   The purpose of a temporary breakpoint is to allow "run to
     *   cursor"-type operations, where the debugger must secretly set its
     *   own internal breakpoint to facilitate the operation but where the
     *   user doesn't intend to add a new breakpoint to the persistent
     *   breakpoint list.  Because the breakpoint is temporary, we won't
     *   bother giving it a meaningful name for the display list, and we
     *   also won't show the breakpoint in the line status displayed in
     *   the source window; this breakpoint should thus be removed the
     *   next time the debugger is entered.
     *   
     *   Only one temporary breakpoint is allowed at a time.  If there's
     *   already a temporary breakpoint set, we'll remove the old one
     *   before adding the new one.  
     */
    int set_temp_bp(struct dbgcxdef *ctx, class CHtmlSysWin *win);

    /*
     *   Clear the temporary breakpoint.  Has no effect if there isn't a
     *   temporary breakpoint.  This should be called each time the
     *   debugger is entered. 
     */
    void clear_temp_bp(struct dbgcxdef *ctx);

    /*
     *   Add a new global breakpoint.  Returns a non-zero error code if
     *   the condition is invalid or any other error occurs; returns zero
     *   on success, and fills in *bpnum with the identifier of the new
     *   breakpoint.  
     */
    int set_global_breakpoint(struct dbgcxdef *ctx, const textchar_t *cond,
                              int *bpnum);

    /*
     *   Enable or disable the breakpoint at the current selection in the
     *   given window.  Does nothing if there's no breakpoint at the given
     *   location. 
     */
    void toggle_bp_disable(struct dbgcxdef *ctx, class CHtmlSysWin *win);

    /*
     *   Delete a breakpoint given the breakpoint number.  The breakpoint
     *   number is obtained from the TADS debugger core; for example, this
     *   is the identifier passed to the enumerator callback function.
     *   Returns zero on success, non-zero if an error occurs.  
     */
    int delete_breakpoint(struct dbgcxdef *ctx, int bpnum);

    /*
     *   Delete all breakpoints 
     */
    void delete_all_bps(struct dbgcxdef *ctx);

    /*
     *   Enable or disable a breakpoint given the breakpoint number. 
     */
    int enable_breakpoint(struct dbgcxdef *ctx, int bpnum, int enable);

    /*
     *   Determine if the given breakpoint number is associated with code.
     *   Returns true if so, false if not.  (Global breakpoints are not
     *   associated with code.) 
     */
    int is_bp_at_line(struct dbgcxdef *ctx, int bpnum);

    /*
     *   Determine if the given breakpoint is a global breakpoint.  Any
     *   breakpoint that doesn't have associated source code is a global
     *   breakpoint. 
     */
    int is_bp_global(struct dbgcxdef *ctx, int bpnum)
        { return !is_bp_at_line(ctx, bpnum); }

    /*
     *   Show the code associated with a given breakpoint.  Returns the
     *   window containing the given code, or null if the breakpoint
     *   doesn't have code associated with it.  
     */
    class CHtmlSysWin *show_bp_code(struct dbgcxdef *ctx, int bpnum,
                                    class CHtmlDebugSysIfc_win *syswinifc);

    /*
     *   Enumerate breakpoints in a displayable format
     */
    void enum_breakpoints(struct dbgcxdef *ctx,
                          void (*cbfunc)(void *cbctx, int bpnum,
                                         const textchar_t *desc,
                                         int disabled),
                          void *cbctx);

    /*
     *   Get a breakpoint's description in a format suitable for display.
     *   Returns zero on success, non-zero on failure. 
     */
    int get_bp_desc(struct dbgcxdef *ctx, int bpnum,
                    textchar_t *buf, size_t buflen);

    /*
     *   Determine if a breakpoint is enabled or disabled.  Returns true
     *   if the breakpoint is enabled, false if it's disabled.  
     */
    int is_bp_enabled(struct dbgcxdef *ctx, int bpnum);

    /*
     *   Get the condition associated with a breakpoint.  Returns zero on
     *   success, non-zero on failure. 
     */
    int get_bp_cond(struct dbgcxdef *ctx, int bpnum,
                    textchar_t *condbuf, size_t condbuflen);

    /*
     *   Change the condition associated with a breakpoint.  Returns zero
     *   on success, non-zero if an error occurred. 
     */
    int change_bp_cond(dbgcxdef *ctx, int bpnum,
                       const textchar_t *cond);

    /*
     *   Save the breakpoint configuration to a configuration object, and
     *   reload a saved configuration 
     */
    void save_bp_config(struct dbgcxdef *ctx, class CHtmlDebugConfig *config);
    void load_bp_config(struct dbgcxdef *ctx, class CHtmlDebugConfig *config);

    /*
     *   Get the text of an error message.  This routine doesn't require
     *   an error stack, so it's suitable for retrieving the text of error
     *   messages returned from the debugger API.  
     */
    void get_error_text(struct dbgcxdef *ctx, int err,
                        textchar_t *buf, size_t buflen);

    /*
     *   Get the text of an error message, formatting the message with
     *   arguments from the error stack.  This should only be used when a
     *   valid error stack is present, such as when entering the debugger
     *   due to a run-time error.  
     */
    void format_error(struct dbgcxdef *ctx, int err, textchar_t *buf,
                      size_t buflen);

    /*
     *   Evaluate an expression in the current stack context.  Fills in
     *   the buffer with the expression result.  If an error occurs, we'll
     *   write an error message to the buffer instead of the expression
     *   value, and we'll also return the error code.  If is_lval is not
     *   null, we'll determine whether the expression can be assigned
     *   into, and set *is_lval to true if so, false if not.  If
     *   is_openable is not null, we'll determine whether the expression
     *   refers to an aggregate value (such as an object or a list), and
     *   set *is_openable to true if so, false if not.
     *   
     *   If aggcb is not null, it refers to an aggregation callback that
     *   we'll invoke for each subitem in the value, if the value refers
     *   to an aggregate object.  The 'relationship' parameter to the
     *   callback gives an operator that can be put between the parent
     *   expression (i.e., the one we're evaluating here) and the subitem
     *   string to yield an expression for the subitem itself.
     *   
     *   If 'speculative' is true, we'll perform speculative evaluation:
     *   we won't allow any assignments, function calls, or method calls.
     *   The idea is to prohibit game state changes, thereby allowing
     *   evaluation of an expression that the user *might* be interested
     *   in before the user actually asks us to evaluate it.  Since the
     *   user hasn't expressly requested evaluation, we don't want to
     *   change anything, thus the prohibitions.  
     */
    int eval_expr(struct dbgcxdef *ctx, textchar_t *buf, size_t buflen,
                  const textchar_t *expr, int *is_lval, int *is_openable,
                  void (*aggcb)(void *aggctx, const char *subitemname,
                                int subitemnamelen, const char *relationship),
                  void *aggctx, int speculative);

    /*
     *   Enumerate local variables in the current stack context.
     */
    void enum_locals(struct dbgcxdef *ctx,
                     void (*cbfunc)(void *ctx, const textchar_t *lclnam,
                                    size_t lclnamlen), void *cbctx);

    /*
     *   Receive notification that a source window is being closed.  This
     *   must be called by the system-specific code; we'll delete the
     *   structures that we allocated to track the window, including its
     *   formatter and parser instances.  Note that the window object
     *   itself must be deleted by the system-specific code. 
     */
    void on_close_srcwin(class CHtmlSysWin *win);

    /*
     *   Static callback version of on_close_srcwin - this can be used to
     *   set up a callback for objects that aren't aware of the helper
     *   class. 
     */
    static void on_close_srcwin_cb(void *ctx, class CHtmlSysWin *win)
    {
        ((CHtmlDebugHelper *)ctx)->on_close_srcwin(win);
    }

    /* get the line link record for a given source line */
    struct CHtmlDbg_line_link *find_line_link(int line_source_id,
                                              unsigned long seekpos);

    /* 
     *   Get the line link record and window link record for the current
     *   selection in a window.  Returns the file seek location of the
     *   start of the line containing the start of the selection range in
     *   the window.  
     */
    unsigned long find_line_info(class CHtmlSysWin *win,
                                 struct CHtmlDbg_win_link **win_link,
                                 struct CHtmlDbg_line_link **line_link);

    /*
     *   Add/remove a status flag to/from a source line.  This will update
     *   the window containing the line to show the new status visually,
     *   if the line is showing in a window.  
     */
    void set_line_status(int line_source_id, unsigned long seekpos,
                         unsigned int statflag)
    {
        change_line_status(line_source_id, seekpos, statflag, TRUE);
    }
    
    void clear_line_status(int line_source_id, unsigned long seekpos,
                           unsigned int statflag)
    {
        change_line_status(line_source_id, seekpos, statflag, FALSE);
    }

    /*
     *   Forget the current line, updating the display to remove the
     *   current line marker. 
     */
    void forget_current_line();

    /* forget the context line */
    void forget_ctx_line();

    /* 
     *   Create a window to use as the stack window.  The helper class
     *   will keep track of the window, and will update it whenever
     *   update_for_entry() is called. 
     */
    class CHtmlSysWin
        *create_stack_win(class CHtmlDebugSysIfc_win *syswinifc);

    /*
     *   Create a window to use as the debug log window. 
     */
    class CHtmlSysWin
        *create_debuglog_win(class CHtmlDebugSysIfc_win *syswinifc);

    /*
     *   Create a window to use as the history trace window.  The helper
     *   class will keep track of the window, and will update it whenever
     *   update_for_entry() is called. 
     */
    class CHtmlSysWin
        *create_hist_win(class CHtmlDebugSysIfc_win *syswinifc);

    /* get the stack window */
    class CHtmlSysWin *get_stack_win() const;

    /* get the debug log window */
    class CHtmlSysWin *get_debuglog_win() const;

    /* get the history window */
    class CHtmlSysWin *get_hist_win() const;

    /* 
     *   Update the stack window If reset_level is true, we'll reset the
     *   context level to the current line; otherwise, we'll refresh the
     *   window and leave the current context line intact. 
     */
    void update_stack_window(struct dbgcxdef *ctx, int reset_level);

    /* update the history window */
    void update_hist_window(struct dbgcxdef *ctx);

    /*
     *   Activate a stack window item.  The stack window should call this
     *   when the user selects a display item in the stack window (in the
     *   manner appropriate to the UI; for example, by double-clicking on
     *   the item).  Returns the source window to be activated.  
     */
    class CHtmlSysWin
        *activate_stack_win_item(class CHtmlDisp *disp, struct dbgcxdef *ctx,
                                 class CHtmlDebugSysIfc_win *syswinifc);

    /*
     *   Activate a stack window item, given the line number.  
     */
    class CHtmlSysWin
        *activate_stack_win_line(int linenum, struct dbgcxdef *ctx,
                                 class CHtmlDebugSysIfc_win *syswinifc);

    /*
     *   Toggle hidden display, and get the current hidden display status
     *   (true -> hidden output is being displayed).
     */
    void toggle_dbg_hid();
    int get_dbg_hid();

    /*
     *   Enumerate open source windows.  Invokes the callback for each
     *   open window displaying either a line source or a text file.  The
     *   'idx' value can be used to identify the window in a subsequent
     *   call to get_source_window_by_index().  
     */
    void enum_source_windows(void (*func)(void *ctx, int idx,
                                          class CHtmlSysWin *win,
                                          int line_source_id,
                                          HtmlDbg_wintype_t win_type),
                             void *cbctx);

    /*
     *   Retrieve a particular source window, given its index.  The index
     *   reflects the same ordering as enum_source_windows() uses; the
     *   first window that enum_source_windows() returns is window 0, the
     *   second is 1, and so on.  
     */
    class CHtmlSysWin *get_source_window_by_index(int idx) const;

    /*
     *   Set the tab expansion size - this sets the number of spaces
     *   between tab stops.  We'll use this setting when loading source
     *   files when we encounter tab (\t, ascii 9) characters.  If there
     *   are open source windows, we'll reload them, since we translate
     *   tabs at load time.  
     */
    void set_tab_size(struct dbgcxdef *ctx,
                      class CHtmlDebugSysIfc_win *win, int n);
    int get_tab_size() const { return tab_size_; }

    /*
     *   Reformat all source windows.  This should be called when there's
     *   a visible change in the display preferences, such as the text
     *   color or font.  
     */
    void reformat_source_windows(struct dbgcxdef *ctx,
                                 class CHtmlDebugSysIfc_win *syswinifc);

    /*
     *   Open a line source, given the line source ID.  Returns the window
     *   containing the line source.  
     */
    class CHtmlSysWin
        *open_line_source(struct dbgcxdef *ctx,
                          class CHtmlDebugSysIfc_win *syswinifc,
                          int line_source_id);

    /*
     *   Open a text file.  We'll try to find an existing window with
     *   showing the same file, and show it if we can find it.  
     */
    class CHtmlSysWin
        *open_text_file(struct dbgcxdef *ctx,
                        class CHtmlDebugSysIfc_win *syswinifc,
                        const char *fname);

    /*
     *   Enumerate the source files making up the game.  Invokes the
     *   callback for each file.  The line source ID is a number that can
     *   be used to identify the line source in operations such as opening
     *   the file; the line sources are numbered consecutively starting at
     *   zero.  
     */
    void enum_source_files(struct dbgcxdef *dbgctx,
                           void (*func)(void *cbctx, const textchar_t *fname,
                                        int line_source_id), void *cbctx);

private:
    /*
     *   Reformat a source file 
     */
    void reformat_source_window(struct dbgcxdef *ctx,
                                class CHtmlDebugSysIfc_win *syswinifc,
                                struct CHtmlDbg_win_link *link);
    
    /* 
     *   reload all source windows -- this must be used when there's a
     *   change in the preferences that affects file loading, such as the
     *   tab expansion size 
     */
    void reload_source_windows(struct dbgcxdef *ctx,
                               class CHtmlDebugSysIfc_win *syswinifc);

    /* change a line status bit */
    void change_line_status(int line_source_id, unsigned long seekpos,
                            unsigned int statflag, int set);

    /* 
     *   Get the line information for a given breakpoint number.  Returns
     *   true if we found a line for the breakpoint, false if the
     *   breakpoint doesn't have a valid line (which is the case for a
     *   global breakpoint). 
     */
    int get_bp_line_info(struct dbgcxdef *dbgctx, int bpnum,
                         int *line_source_id, unsigned long *seekpos);

    /* update all lines on screen for a particular source window */
    void update_lines_on_screen(struct CHtmlDbg_win_link *win_link);

    /* update a line's status on-screen */
    void update_line_on_screen(struct CHtmlDbg_line_link *line_link);

    /* update a line's status in a particular window */
    void update_line_on_screen(struct CHtmlDbg_win_link *win_link,
                               struct CHtmlDbg_line_link *line_link);

    /* display callback for updating the stack window */
    static void update_stack_disp_cb(void *ctx, char *str, int strl);
    void update_stack_disp(struct HtmlDbg_stkdisp_t *ctx,
                           char *str, int strl);

    /* breakpoint enumeration callback for saving breakpoint configuration */
    static void save_bp_enumcb(void *ctx0, int bpnum,
                               const char *desc, const char *cond,
                               int disabled);

    /* breakpoint enumeration callback for generating displayable bp list */
    static void disp_bp_enumcb(void *ctx0, int bpnum,
                               const char *desc, const char *cond,
                               int disabled);

    /*
     *   Format a breakpoint description from our internal format to a
     *   format suitable for display to the user.  
     */
    static void format_bp_for_display(textchar_t *buf, size_t buflen,
                                      const textchar_t *privatedesc,
                                      const textchar_t *cond, int disabled);

    /*
     *   Load the source file for the current line (i.e., the point at
     *   which execution is currently suspended), if it's not already
     *   loaded, and return the (possibly new) window showing that source
     *   file.  
     */
    class CHtmlSysWin
        *load_current_source(struct dbgcxdef *ctx,
                             class CHtmlDebugSysIfc_win *sysifc)
    {
        /* load the source at stack level 0 */
        return load_source_at_level(ctx, sysifc, 0);
    }

    /*
     *   Load the source file for the current line at a given stack level.
     *   Level 0 is the current frame, 1 is the first enclosing frame, and
     *   so on. 
     */
    class CHtmlSysWin
        *load_source_at_level(struct dbgcxdef *ctx,
                              class CHtmlDebugSysIfc_win *sysifc, int level);

    /*
     *   Load a source file into a new window and return the window.  If
     *   the file has already been loaded into a window, this simply
     *   return the existing window.  
     */
    struct CHtmlDbg_win_link *
        load_source_file(struct dbgcxdef *ctx,
                         class CHtmlDebugSysIfc_win *sysifc,
                         struct linfdef *linf, int need_activate);

    /*
     *   Load a source file into a window 
     */
    void load_file_into_win(struct dbgcxdef *ctx,
                            struct CHtmlDbg_win_link *win_link,
                            struct linfdef *linf, int need_activate);

    /*
     *   create a new source window; returns the tracking structure for
     *   the new window 
     */
    struct CHtmlDbg_win_link
        *create_source_win(class CHtmlDebugSysIfc_win *syswinifc,
                           struct linfdef *linf);

    /* create a debugger window (source or other) */
    struct CHtmlDbg_win_link
        *create_win(class CHtmlDebugSysIfc_win *syswinifc,
                    struct linfdef *linf, const textchar_t *win_title,
                    HtmlDbg_wintype_t win_type);

    /* add a window to/remove a window from our source window list */
    struct CHtmlDbg_win_link *add_src_win(class CHtmlSysWin *win,
                                          class CHtmlParser *parser,
                                          class CHtmlFormatter *formatter,
                                          struct linfdef *linf,
                                          const textchar_t *win_title,
                                          class CHtmlSysFont *font,
                                          HtmlDbg_wintype_t win_type);
    void del_src_win(struct CHtmlDbg_win_link *link);

    /* display callback for eval_expr */
    static void eval_expr_cb(void *ctx, char *txt, int len);

    /* error logging callback for eval_expr */
    static void eval_expr_errlog(void *ctx0, char * /*fac*/, int err,
                                 int argc, union erradef *argv);

    /* 
     *   Get the object and p-code offset for the current selection in the
     *   given window.  Returns zero on success, non-zero on failure.  
     */
    int srcofs_to_code(struct dbgcxdef *ctx, class CHtmlSysWin *win,
                       unsigned long *fpos,
                       struct CHtmlDbg_win_link **win_link,
                       objnum *line_objn, uint *line_ofs);

    /* head and tail of source window list */
    struct CHtmlDbg_win_link *first_srcwin_;
    struct CHtmlDbg_win_link *last_srcwin_;

    /* head and tail of special line list */
    struct CHtmlDbg_line_link *first_line_;
    struct CHtmlDbg_line_link *last_line_;

    /* currently-executing line */
    int cur_is_valid_ : 1;
    int cur_source_id_;
    unsigned long cur_seekpos_;

    /* 
     *   Current context line - this is the line that's selected as the
     *   current evaluation context from the stack list.  If this is not
     *   valid, the current line is used instead. 
     */
    int ctx_is_valid_ : 1;
    int ctx_source_id_;
    unsigned long ctx_seekpos_;

    /* display item in stack window for current evaluation context level */
    class CHtmlDispTextDebugsrc *ctx_stackwin_disp_;

    /* the stack window */
    struct CHtmlDbg_win_link *stack_win_;

    /* the debug log window */
    struct CHtmlDbg_win_link *debuglog_win_;

    /* the history window */
    struct CHtmlDbg_win_link *hist_win_;

    /* tab expansion size setting */
    int tab_size_;

    /* temporary breakpoint number, and flag indicating if it's set */
    int tmp_bpnum_;
    int tmp_bp_valid_ : 1;
};


/* ------------------------------------------------------------------------ */
/*
 *   System window tracker.  We keep a list of the source windows we've
 *   opened using a linked list of these structures. 
 */
struct CHtmlDbg_win_link
{
    CHtmlDbg_win_link(class CHtmlSysWin *win, class CHtmlParser *parser,
                      class CHtmlFormatter *formatter, struct linfdef *linf,
                      const textchar_t *win_title,
                      class CHtmlSysFont *font, HtmlDbg_wintype_t win_type);

    /* 
     *   find the line source in the debug context for the given line
     *   source ID 
     */
    static struct lindef *find_lindef(struct dbgcxdef *ctx, int id);

    /* determine if I match a given linfdef */
    int linf_equals(struct linfdef *linf);

    /* determine if my filename is the same as that of a linfdef */
    int fname_equals(struct linfdef *linf);

    /* next and previous link in the list */
    CHtmlDbg_win_link *next_;
    CHtmlDbg_win_link *prev_;

    /* our system window */
    class CHtmlSysWin *win_;

    /* the parser associated with the window */
    class CHtmlParser *parser_;

    /* formatter associated with the window */
    class CHtmlFormatter *formatter_;

    /* name of the file */
    CStringBuf fname_;

    /* line source ID */
    int source_id_;

    /* default font */
    class CHtmlSysFont *font_;

    /* type of window */
    HtmlDbg_wintype_t win_type_;
};

/* ------------------------------------------------------------------------ */
/*
 *   Special-line tracker.  We keep a list of lines that have a special
 *   characteristic, such as being the current line executing or having a
 *   breakpoint.  These characteristics show up visually in source
 *   windows, so we need to update the corresponding source window
 *   whenever a line's status changes.  Since a line might not always be
 *   loaded into a window when its status changes, this structure lets us
 *   update a window when it's initially loaded, as well as keeping it up
 *   to date while it's loaded. 
 */

struct CHtmlDbg_line_link
{
    CHtmlDbg_line_link(int source_id, unsigned long seekpos,
                       unsigned int init_stat);

    /* determine if this line matches a given line */
    int equals(int source_id, unsigned long seekpos)
    {
        return (source_id_ == source_id && seekpos_ == seekpos);
    }
    
    /* next/previous link in the list */
    CHtmlDbg_line_link *next_;
    CHtmlDbg_line_link *prev_;
    
    /* 
     *   Line source ID (linid) of the file containing the line.  Since
     *   special lines are always source files referenced by the .GAM file
     *   (we can't have breakpoints or execution in any files not in the
     *   debug records, after all), special lines always have valid line
     *   source ID's. 
     */
    int source_id_;

    /* seek location of the line */
    unsigned long seekpos_;

    /* 
     *   The current set of status flags associated with the line.  If a
     *   line's status flags become zero, we'll drop the link from the
     *   list (after updating the screen accordingly), since there's
     *   nothing to keep track of any more. 
     */
    unsigned int stat_;
};


#endif /* HTMLDBG_H */

