/* $Header: d:/cvsroot/tads/html/htmldcfg.h,v 1.1.1.1 1999/05/09 21:32:28 MJRoberts Exp $ */

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  htmldcfg.h - debugger configuration reader/writer
Function
  This is a simple utility object that allows saving and restoring
  debugger state information.

  Information is stored as a set of name/value pairs.  Each value can
  be a string, list of strings, rectangle, or boolean value.  Each name
  can have a single associated value, or can have an array of associated
  values.  Array items are addressed by an element name (rather than
  an index number).

  The configuration information is kept in memory while running.  The
  information can be loaded from a file or saved to a file at any time;
  normally, a configuration file would be loaded at startup and written
  at program exit.
Notes
  
Modified
  03/05/98 MJRoberts  - Creation
*/

#ifndef HTMLDCFG_H
#define HTMLDCFG_H

#ifndef TADSHTML_H
#include "tadshtml.h"
#endif


/* maximum length of a key name */
const int HTML_DCFG_KEY_MAX = 512;

class CHtmlDebugConfig
{
public:
    CHtmlDebugConfig();
    ~CHtmlDebugConfig();

    /* 
     *   Load a file; replaces all current configuration information.
     *   Returns zero on success, non-zero no failure. 
     */
    int load_file(const textchar_t *fname);

    /* 
     *   Write a file.  Returns zero on success, non-zero on failure. 
     */
    int save_file(const textchar_t *fname);

    /*
     *   Store values.  The variable name and the element ID together form
     *   the key; the element ID can be null.  If no variable of the given
     *   name exists, one is created.  If a variable of this name already
     *   exists, its value is replaced.  
     */
    void setval(const textchar_t *varname, const textchar_t *element,
                int val);
    void setval(const textchar_t *varname, const textchar_t *element,
                int stridx, const textchar_t *val);
    void setval(const textchar_t *varname, const textchar_t *element,
                int stridx, const textchar_t *val, size_t vallen);
    void setval(const textchar_t *varname, const textchar_t *element,
                const CHtmlRect *val);
    void setval_color(const textchar_t *varname, const textchar_t *element,
                      HTML_color_t color);
    void setval_bytes(const textchar_t *varname, const textchar_t *element,
                      const void *buf, size_t len);

    /* clear a string list - removes all elements from the list */
    void clear_strlist(const textchar_t *varname, const textchar_t *element);

    /* get the number of elements in a string list */
    int get_strlist_cnt(const textchar_t *varname, const textchar_t *element);

    /* 
     *   Retrieve values.  These routines return zero on success, non-zero
     *   on failure.  The call fails if the named value does not exist or
     *   is not of the requested type.  
     */
    int getval(const textchar_t *varname, const textchar_t *element,
               int *val);
    int getval(const textchar_t *varname, const textchar_t *element,
               int stridx, textchar_t *dstbuf, size_t dstbuflen)
    {
        return getval(varname, element, stridx, dstbuf, dstbuflen, 0);
    }
    
    int getval(const textchar_t *varname, const textchar_t *element,
               int stridx, textchar_t *dstbuf, size_t dstbuflen,
               size_t *result_len);
    int getval(const textchar_t *varname, const textchar_t *element,
               CHtmlRect *val);
    int getval_color(const textchar_t *varname, const textchar_t *element,
                     HTML_color_t *color);
    
    int getval_bytes(const textchar_t *varname, const textchar_t *element,
                     void *buf, size_t buflen);
    int getval_bytes_len(const textchar_t *varname,
                         const textchar_t *element, size_t *len);
    int getval_bytes_buf(const textchar_t *varname,
                         const textchar_t *element, void **bufptr);

protected:
    /* hash table entry enumeration callback for saving to a file */
    static void save_enumcb(void *ctx, class CHtmlHashEntry *entry);

    /* 
     *   Get the entry with a given hash key.  If create is true, we'll
     *   create a new entry if one doesn't exist.
     */
    class CHtmlHashEntryConfig *get_entry(const textchar_t *varname,
                                          const textchar_t *element,
                                          int create);

    /* hash table containing the values */
    class CHtmlHashTable *hashtab_;

    /* file signature */
    static const textchar_t filesig[];
};


#endif /* HTMLDCFG_H */

