#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/html/htmlrf.cpp,v 1.1.1.1 1999/05/09 21:32:27 MJRoberts Exp $";
#endif

/* Copyright (c) 1997 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  htmlrf.cpp - HTML resource finder
Function
  
Notes
  
Modified
  12/03/97 MJRoberts  - Creation
*/

/* include TADS os interface functions */
#include <os.h>

#ifndef HTMLRF_H
#include "htmlrf.h"
#endif
#ifndef HTMLHASH_H
#include "htmlhash.h"
#endif

/* ------------------------------------------------------------------------ */
/*
 *   Hash table entry subclass for a resource map entry
 */
class CHtmlHashEntryFileres: public CHtmlHashEntryCI
{
public:
    CHtmlHashEntryFileres(const textchar_t *str, size_t len,
                          unsigned long seekofs, unsigned long siz,
                          int fileno)
        : CHtmlHashEntryCI(str, len, TRUE)
    {
        seekofs_ = seekofs;
        siz_ = siz;
        fileno_ = fileno;
    }

    /* seek offset and size of the resource in the resource file */
    unsigned long seekofs_;
    unsigned long siz_;

    /* 
     *   File number - the resource finder keeps a list of resource files
     *   that we've mapped; this gives the ID of the file in the list.  
     */
    int fileno_;
};

/* ------------------------------------------------------------------------ */
/*
 *   File list entry.  The resource finder keeps a list of files that
 *   we've loaded.  
 */
class CHtmlRfFile
{
public:
    CHtmlRfFile() { }
    CHtmlRfFile(const textchar_t *fname)
    {
        filename_.set(fname);
    }

    /* get the filename */
    const textchar_t *get_fname() { return filename_.get(); }

    /* change the filename */
    void set_fname(const textchar_t *fname) { filename_.set(fname); }

    /* get/set the seek base position */
    unsigned long get_seek_base() const { return res_seek_base_; }
    void set_seek_base(unsigned long val) { res_seek_base_ = val; }

private:
    /* name of the file */
    CStringBuf filename_;

    /* base seek offset of resources in .GAM file */
    unsigned long res_seek_base_;
};

/* ------------------------------------------------------------------------ */
/*
 *   Resource finder implementation 
 */

CHtmlResFinder::CHtmlResFinder()
{
    /* create a hash table for the resource map */
    restab_ = new CHtmlHashTable(256, new CHtmlHashFuncCI);

    /* 
     *   Initialize the file list.  We always allocate the first entry,
     *   since this is the dedicated entry for the .GAM file. 
     */
    files_alloced_ = 5;
    files_ = (CHtmlRfFile **)
             th_malloc(files_alloced_ * sizeof(CHtmlRfFile *));
    files_used_ = 1;
    files_[0] = new CHtmlRfFile();
}

CHtmlResFinder::~CHtmlResFinder()
{
    int i;
    
    /* delete the hash table */
    delete restab_;

    /* delete the list of files */
    for (i = 0 ; i < files_used_ ; ++i)
        delete files_[i];
    delete files_;
}


/*
 *   Get the filename and seek offset to use to read a resource.  If we
 *   can find the resource in our .GAM file resource map, we'll return the
 *.  GAM filename and the offset of the resource in the
 *.  GAM file; otherwise, we'll return the name of the resource itself
 *   and a seek offset of zero.  
 */
void CHtmlResFinder::get_file_info(CStringBuf *outbuf,
                                   const textchar_t *resname,
                                   size_t resnamelen,
                                   unsigned long *start_seekpos,
                                   unsigned long *file_size)
{
    CHtmlHashEntryFileres *entry;
    
    /* see if we can find an entry in the resource map */
    entry = (CHtmlHashEntryFileres *)restab_->find(resname, resnamelen);
    if (entry != 0)
    {
        /* 
         *   Found an entry - return the seek position of the entry in the
         *   file (calculate the actual seek position as the resource map
         *   base seek address plus the offset of this resource within the
         *   resource map).  Get the filename from the files_[] list entry
         *   for the file number stored in the resource entry.  
         */
        *start_seekpos = files_[entry->fileno_]->get_seek_base()
                         + entry->seekofs_;
        *file_size = entry->siz_;
        outbuf->set(files_[entry->fileno_]->get_fname());
    }
    else
    {
        char cvtbuf[OSFNMAX];
        char *root;
        osfildef *fp;
        
        /*
         *   There's no entry in the resource map matching this name - map
         *   the resource name to a filename, and start at offset zero in
         *   the file.  Prefix the resource path with the path to the
         *   original game file, which we always store in files_[0].  
         */
        *start_seekpos = 0;
        strcpy(cvtbuf, files_[0]->get_fname());
        root = os_get_root_name(cvtbuf);

        /* add the resource name after the path to the game file */
        os_cvt_url_dir(root, sizeof(cvtbuf) - (root - cvtbuf),
                       resname, FALSE);
        outbuf->set(cvtbuf);

        /* determine the size of the file */
        fp = osfoprb(cvtbuf, OSFTBIN);
        if (fp == 0)
        {
            /* no file */
            *file_size = 0;

            /* log a debug message */
            oshtml_dbg_printf("resource loader error: resource not found: "
                              " %s\n", resname);
        }
        else
        {
            /* seek to the end of the file to determine its size */
            osfseek(fp, 0, OSFSK_END);
            *file_size = (unsigned long)osfpos(fp);

            /* done with the file */
            osfcls(fp);
        }
    }
}

/*
 *   Set the name of the game file 
 */
void CHtmlResFinder::set_game_name(const char *fname)
{
    /* 
     *   Remember the name of the game file - we'll use this to look up
     *   resources.  The first entry in our file list is dedicated to the
     *   GAM file, so set this entry's filename.  In addition, we'll use
     *   the path to the game when looking for resources, since we look in
     *   the game's directory for external resources.  
     */
    files_[0]->set_fname(fname);
}

/*
 *   Add a resoure.  We'll add an entry to our hash table with the
 *   resource information.
 */
void CHtmlResFinder::add_resource(unsigned long ofs, unsigned long siz,
                                  const char *nm, size_t nmlen, int fileno)
{
    /* create and add a new hash table entry */
    restab_->add(new CHtmlHashEntryFileres(nm, nmlen, ofs, siz, fileno));
}

/*
 *   Set the resource seek base for a resource file 
 */
void CHtmlResFinder::set_resmap_seek(unsigned long seekpos, int fileno)
{
    files_[fileno]->set_seek_base(seekpos);
}

/*
 *   Add a resource file, returning the file number for the new resource
 *   file.  
 */
int CHtmlResFinder::add_resfile(const char *fname)
{
    int fileno;
    
    /* note the new file's index */
    fileno = files_used_;
    
    /*
     *   If there isn't room in our array of resource files, expand the
     *   array.  
     */
    if (files_used_ == files_alloced_)
    {
        files_alloced_ += 5;
        files_ = (CHtmlRfFile **)
                 th_realloc(files_, files_alloced_ * sizeof(CHtmlRfFile *));
    }

    /* add the new entry */
    files_[files_used_] = new CHtmlRfFile(fname);
    ++files_used_;

    /* return the new file number */
    return fileno;
}


/* ------------------------------------------------------------------------ */
/*
 *   Static functions invoked from TADS loader.  We'll refer these to an
 *   instance of a CHtmlResFinder, which is stored in the callback
 *   context. 
 */

void CHtmlResFinder::cb_set_game_name(void *ctx, const char *fname)
{
    ((CHtmlResFinder *)ctx)->set_game_name(fname);
}

void CHtmlResFinder::cb_set_resmap_seek(void *ctx, unsigned long seekpos,
                                        int fileno)
{
    ((CHtmlResFinder *)ctx)->set_resmap_seek(seekpos, fileno);
}

void CHtmlResFinder::cb_add_resource(void *ctx, unsigned long ofs,
                                     unsigned long siz,
                                     const char *nm, size_t nmlen,
                                     int fileno)
{
    ((CHtmlResFinder *)ctx)->add_resource(ofs, siz, nm, nmlen, fileno);
}

int CHtmlResFinder::cb_add_resfile(void *ctx, const char *fname)
{
    return ((CHtmlResFinder *)ctx)->add_resfile(fname);
}
