#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/html/win32/tadswav.cpp,v 1.1.1.1 1999/05/09 21:32:31 MJRoberts Exp $";
#endif

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  tadswav.cpp - TADS WAV (waveform audio) file support
Function
  
Notes
  
Modified
  01/19/98 MJRoberts  - Creation
*/

#include <Windows.h>
#include <dsound.h>

#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <process.h>

/* include TADS OS layer for I/O functions */
#include <os.h>

#ifndef TADSHTML_H
#include "tadshtml.h"
#endif
#ifndef TADSWAV_H
#include "tadswav.h"
#endif


/* ------------------------------------------------------------------------ */
/*
 *   Wave file player implementation 
 */

CTadsWavFilePlayer::CTadsWavFilePlayer()
{
    /* clear members */
    reader_ = 0;
    istreambuf_ = 0;
    inotify_ = 0;
    notify_evt_[0] = notify_evt_[1] = 0;
    thread_hdl_ = 0;
    idirsnd_ = 0;
    prepared_ = FALSE;

    /* initialize our critical section object */
    InitializeCriticalSection(&sync_);
}

CTadsWavFilePlayer::~CTadsWavFilePlayer()
{
    /* delete anything we've created */
    clean_up();

    /* release our critical section object */
    DeleteCriticalSection(&sync_);
}

/*
 *   clean up any objects we've created 
 */
void CTadsWavFilePlayer::clean_up()
{
    /* if we're still running, stop playback */
    if (thread_hdl_ != 0)
        stop(TRUE);

    /* if we have a reader, delete it */
    if (reader_ != 0)
    {
        delete reader_;
        reader_ = 0;
    }

    /* 
     *   done with the notification thread resources - this release the
     *   DirectSound interfaces and the notification event handles 
     */
    delete_notify_resources();

    /* no longer prepared */
    prepared_ = FALSE;
}

/*
 *   Release the resources that the notification thread uses.  We call
 *   this from the notification thread when it's about to exit, and we can
 *   also call it from the destructor if the notification thread never
 *   started.  
 */
void CTadsWavFilePlayer::delete_notify_resources()
{
    int i;

    /* this routine must be synchronized across threads */
    EnterCriticalSection(&sync_);

    /* release our stream buffer, if we have one */
    if (istreambuf_ != 0)
    {
        istreambuf_->Release();
        istreambuf_ = 0;
    }

    /* release our notification interface, if we have one */
    if (inotify_ != 0)
    {
        inotify_->Release();
        inotify_ = 0;
    }

    /* release our direct sound main interface */
    if (idirsnd_ != 0)
    {
        idirsnd_->Release();
        idirsnd_ = 0;
    }

    /* delete notification event handles */
    for (i = 0 ; i < sizeof(notify_evt_)/sizeof(notify_evt_[0]) ; ++i)
    {
        if (notify_evt_[i] != 0)
        {
            CloseHandle(notify_evt_[i]);
            notify_evt_[i] = 0;
        }
    }

    /* no longer prepared */
    prepared_ = FALSE;

    /* end of critical section */
    LeaveCriticalSection(&sync_);
}

/*
 *   Prepare for playback.  This allocates all resources and queues up the
 *   WAV file at the start of the data.  Returns zero on success, non-zero
 *   on failure.  If we're already prepared, we'll return success.  
 */
int CTadsWavFilePlayer::prepare(IDirectSound *ids, HANDLE hfile,
                                HWND win_cb, UINT win_cb_msg,
                                LPARAM win_cb_lpar, int repeat)
{
    WAVEFORMATEX *fmt;
    DSBUFFERDESC bufdesc;

    /* if we're already prepared, there's nothing to do */
    if (prepared_)
        return 0;

    /* can't play if there's no DirectSound interface */
    if (ids == 0)
        return 1;

    /* clean up any old state we had */
    clean_up();

    /* note the repeat count */
    repeat_ = repeat;
    repeats_done_ = 0;

    /* remember the callback information */
    win_cb_ = win_cb;
    win_cb_msg_ = win_cb_msg;
    win_cb_lpar_ = win_cb_lpar;

    /* create a file reader and parse the header */
    reader_ = new CTadsWavFileReader();
    if (reader_->read_header(hfile))
    {
        clean_up();
        return 1;
    }

    /*
     *   Calculate the buffer length.  Use a buffer big enough to hold
     *   roughly three seconds of playback.  The size of ach chunk must be
     *   an integral multiple of the block alignment, and the overall
     *   buffer size must be an integral multiple of the chunk size, so
     *   that we can tile the buffer with the chunks.  
     */
    fmt = reader_->get_wavefmtex();
    chunksize_ = (fmt->nSamplesPerSec * 3 * (DWORD)fmt->nBlockAlign)
                 / TADSWAV_NOTIFY_CNT;

    /* 
     *   adjust the chunk size so it divides the total file length as
     *   close to evenly as possible 
     */
    if (reader_->get_wave_len() / chunksize_ > 0)
        chunksize_ = (reader_->get_wave_len()
                      / (reader_->get_wave_len() / chunksize_)) + 1;

    /* align the chunk size on the format block alignment size */
    if (chunksize_ % (DWORD)fmt->nBlockAlign != 0)
        chunksize_ += fmt->nBlockAlign - (chunksize_ % fmt->nBlockAlign);

    /* make the buffer an even multiple of the chunk size */
    bufsize_ = chunksize_ * TADSWAV_NOTIFY_CNT;

    /* set up the buffer descriptor for DirectSound */
    memset(&bufdesc, 0, sizeof(bufdesc));
    bufdesc.dwSize = sizeof(bufdesc);
    bufdesc.dwFlags = DSBCAPS_CTRLPOSITIONNOTIFY
                      | DSBCAPS_CTRLDEFAULT
                      | DSBCAPS_GETCURRENTPOSITION2;
    bufdesc.dwBufferBytes = bufsize_;
    bufdesc.lpwfxFormat = fmt;

    /* 
     *   Remember our DirectSound interface, and set a reference on it -
     *   the sound buffer seems to require that the IDirectSound object
     *   still exist when the buffer is deleted, but strangely the buffer
     *   doesn't set its own reference on the IDirectSound object.  In
     *   order to prevent crashes at buffer deletion time, keep a
     *   reference on the IDirectSound interface until we've released the
     *   buffer.  
     */
    idirsnd_ = ids;
    idirsnd_->AddRef();

    /* create the buffer */
    if (ids->CreateSoundBuffer(&bufdesc, &istreambuf_, 0) != DS_OK)
    {
        clean_up();
        return 2;
    }

    /* we're not done reading or playing yet */
    done_reading_ = done_playing_ = FALSE;

    /* get the notification interface */
    if (istreambuf_->QueryInterface(IID_IDirectSoundNotify,
                                    (void **)&inotify_) != DS_OK)
    {
        /* 
         *   can't use notifications - we'll use a timer instead; that's
         *   less precise, so our latency in noticing that the sound is
         *   over will be a little higher, but it will work with older
         *   versions of DirectX (in particular with NT 4, which only
         *   supports DirectX 3) 
         */
        inotify_ = 0;
    }

    /* load the buffer */
    if (load_init_buffer())
    {
        clean_up();
        return 4;
    }

    /* initialize notifications */
    if (init_notify())
    {
        clean_up();
        return 5;
    }

    /* we're now prepared */
    prepared_ = TRUE;

    /* success */
    return 0;
}

/*
 *   Begin playback of a WAV file.  If anything goes wrong, we'll clean up
 *   and return a non-zero error indicator; on success, we'll return zero,
 *   and the sound will be playing asynchronously.  
 */
int CTadsWavFilePlayer::play(IDirectSound *ids, HANDLE hfile,
                             HWND win_cb, UINT win_cb_msg, LPARAM win_cb_lpar,
                             int repeat)
{
    int err;

    /* prepare for playback */
    if ((err = prepare(ids, hfile, win_cb,
                       win_cb_msg, win_cb_lpar, repeat)) != 0)
        return err;

    /* we haven't played back any bytes yet */
    bytes_played_ = 0;

    /* begin playback */
    if (istreambuf_->Play(0, 0, DSBPLAY_LOOPING) != DS_OK)
    {
        clean_up();
        return 6;
    }

    /* if we don't have a notifier, tell the timer thread to start */
    if (inotify_ == 0)
        SetEvent(notify_evt_[0]);

    /* success */
    return 0;
}

/*
 *   stop a WAV file 
 */
void CTadsWavFilePlayer::stop(int sync)
{
    /* if we don't have a stream buffer, there's nothing to do */
    if (istreambuf_ == 0)
        return;
    
    /* tell the stream buffer to stop */
    istreambuf_->Stop();

    /* 
     *   if we're not using notifiers, trigger the notifier event object
     *   to tell the playback thread that we're done 
     */
    if (inotify_ == 0)
        SetEvent(notify_evt_[1]);

    /* 
     *   if this is synchronous, wait for the notification service thread
     *   to exit; this won't happen until playback has actually stopped 
     */
    if (sync)
    {
        /* wait for our stop event, allowing a 5-second timeout */
        WaitForSingleObject(thread_hdl_, 5000);

        /* we can close the thread handle now */
        CloseHandle(thread_hdl_);
        thread_hdl_ = 0;
    }
}


/*
 *   Load the initial buffer.  This is used before playback begins to fill
 *   the buffer with the first set of data from the file.  
 */
int CTadsWavFilePlayer::load_init_buffer()
{
    BYTE *write1, *write2;
    DWORD len1, len2;
    unsigned long bytes_read;

    /* lock the entire buffer, since we want to fill it all */
    if (istreambuf_->Lock(0, bufsize_, (void **)&write1, &len1,
                          (void **)&write2, &len2, 0) != DS_OK
        || len2 != 0)
        return 1;

    /* seek to the start of the data section */
    reader_->seek_data_start();

    /* load the buffer */
    if (reader_->read_data(write1, len1, &bytes_read,
                           repeat_, &repeats_done_))
        return 2;

    /* unlock the buffer */
    if (istreambuf_->Unlock(write1, len1, 0, 0) != DS_OK)
        return 3;

    /* if we didn't read the full buffer, we have a short file */
    if (bytes_read < len1)
    {
        /* note that we've read the entire file */
        done_reading_ = TRUE;

        /* 
         *   fill the remainder of the buffer with silence, in case the
         *   hardware play cursor moves past the end of the part we read
         *   before we can cancel playback 
         */
        FillMemory(write1 + bytes_read, len1 - bytes_read,
                   (BYTE)(reader_->get_wavefmtex()
                          ->wBitsPerSample == 8 ? 128 : 0));

        /*
         *   If we're using the DirectX 5 notify scheme, we need to align
         *   everything to the chunk size, so fill out the remainder of
         *   the unread chunk with silence and advance the write position
         *   to the start of the next chunk.  If we're using the older
         *   timer-based scheme, leave everything unaligned. 
         */
        if (inotify_ != 0)
        {
            /* 
             *   using DirectX 5 notifiers, so we need to align everything
             *   on chunk boundaries 
             */
            
            /* remember the offset of the last block that we read */
            last_block_ofs_ = (bytes_read / chunksize_ + 1) * chunksize_;
            if (last_block_ofs_ >= bufsize_)
                last_block_ofs_ = 0;

            /* move the write offset to the start of the next chunk */
            write_ofs_ = len1;
        }
        else
        {
            /* move the write offset to the end of the part we read */
            write_ofs_ = bytes_read;
        }
    }
    else
    {
        /* 
         *   we read the entire buffer, so wrap around and start writing
         *   at the beginning of the buffer 
         */
        write_ofs_ = 0;
    }

    /* wrap the write offset if it moved off the end of the buffer */
    if (write_ofs_ >= bufsize_)
        write_ofs_ -= bufsize_;

    /* reading from the beginning */
    read_ofs_ = 0;

    /* success */
    return 0;
}

/*
 *   Initialize notifications in preparation for playback
 */
int CTadsWavFilePlayer::init_notify()
{
    /* create two events: one for chunk notifications, one for stop */
    notify_evt_[0] = CreateEvent(0, FALSE, FALSE, 0);
    notify_evt_[1] = CreateEvent(0, FALSE, FALSE, 0);
    if (notify_evt_[0] == 0 || notify_evt_[1] == 0)
        return 1;

    /*
     *   If this version of DirectX supports notifications, use those;
     *   otherwise, we'll have to rely on a timer instead. 
     */
    if (inotify_ != 0)
    {
        DSBPOSITIONNOTIFY notifypos[TADSWAV_NOTIFY_CNT + 1];
        int i;
        
        /* set up the event descriptors for the chunk notifiers */
        notifypos[0].dwOffset = chunksize_;
        notifypos[0].hEventNotify = notify_evt_[0];
        for (i = 1 ; i < TADSWAV_NOTIFY_CNT ; ++i)
        {
            notifypos[i].dwOffset = notifypos[i-1].dwOffset + chunksize_;
            notifypos[i].hEventNotify = notify_evt_[0];
        }
        
        /* 
         *   get the last notification at the last byte (rather than one
         *   past the last byte) of the buffer 
         */
        notifypos[i-1].dwOffset--;
        
        /* set up the event descriptor for the STOP notifier */
        notifypos[i].dwOffset = DSBPN_OFFSETSTOP;
        notifypos[i].hEventNotify = notify_evt_[1];
        
        /* create the thread that will handle the notification events */
        thread_hdl_ = CreateThread(0, 0,
                                   (LPTHREAD_START_ROUTINE)notify_thread_main,
                                   (void *)this, 0, &thread_id_);
        if (thread_hdl_ == 0)
            return 2;
        
        /* set up notifications */
        if (inotify_->SetNotificationPositions(TADSWAV_NOTIFY_CNT + 1,
                                               notifypos) != DS_OK)
            return 3;
    }
    else
    {
        /* create the thread that will fill the buffer using a timer */
        thread_hdl_ = CreateThread(0, 0,
                                   (LPTHREAD_START_ROUTINE)timer_thread_main,
                                   (void *)this, 0, &thread_id_);
        if (thread_hdl_ == 0)
            return 10;
    }
        
    /* success */
    return 0;
}

/*
 *   Timer thread main entrypoint
 */
void CTadsWavFilePlayer::timer_thread_main(void *ctx)
{
    /* call the member function to handle the thread */
    ((CTadsWavFilePlayer *)ctx)->do_timer();
}

/*
 *   Timer thread main handler.  We'll try to keep the buffer filled by
 *   waking up a few times a second and reading from the file into the
 *   buffer to fill the portion that's been played since the last time we
 *   woke up.
 *   
 *   write_ofs_ is the next position into which we load bytes; it starts
 *   out at the end of the block initially read from the file, and we'll
 *   advance it to the end of each new block read from the file.
 *   
 *   We'll use write_ofs_ and the current play cursor to determine how
 *   much we can read.  We can always fill behind the current play cursor,
 *   so we'll fill the part from the last write offset to the current play
 *   position.  
 */
DWORD CTadsWavFilePlayer::do_timer()
{
    DWORD playpos, writepos;
    DWORD locklen;
    unsigned long silence_start, silence_end;
    int done;
    BYTE *write1, *write2;
    DWORD len1, len2;
    int err;
    unsigned long bytes_read;
    BYTE silence = (BYTE)(reader_->get_wavefmtex()
                          ->wBitsPerSample == 8 ? 128 : 0);

    /* wait for the startup event */
    switch(WaitForMultipleObjects(2, notify_evt_, FALSE, INFINITE))
    {
    case WAIT_OBJECT_0:
        /* playback started - proceed into main loop */
        break;

    case WAIT_OBJECT_0 + 1:
        /* playback halted - return immediately */
        return 0;

    default:
        /* anything else is an error */
        return 1;
    }
    
    /*
     *   If we've already read the whole file, calculate the location of
     *   the silent part -- since everything fits into one buffer, it's
     *   simply from the write offset to the end of the buffer. 
     */
    if (done_reading_)
    {
        silence_start = write_ofs_;
        silence_end = bufsize_;
    }

    /* 
     *   keep looping until we run out of data to play back or playback is
     *   halted externally 
     */
    for (err = 0, done = FALSE ; !done && err == 0 ; )
    {
        /* 
         *   wait until our timeout has elapsed, or the 'stop' event is
         *   triggered, which tells us that playback has been stopped 
         */
        switch(WaitForSingleObject(notify_evt_[1], 500))
        {
        case WAIT_OBJECT_0:
            /* 
             *   The 'stop' event object has been signalled - this means
             *   that someone has called our stop method, which means that
             *   we're halting playback.  We can immediately terminate
             *   this thread, since we don't need to keep filling the
             *   buffer any more now that playback is going to cease.  
             */
            done = TRUE;
            break;

        case WAIT_TIMEOUT:
            /*
             *   Our time-out interval has elapsed.  We can load from the
             *   last write offset to the current play cursor (we can
             *   always load data behind the play cursor).  First, get the
             *   current buffer information.  
             */
            istreambuf_->GetCurrentPosition(&playpos, &writepos);

            /* 
             *   We can lock from the last write position to the new play
             *   position.  If the current play offset is less than the
             *   old write offset, it means the play cursor has wrapped
             *   around in the buffer.  Calculate the size of the block we
             *   can lock.  
             */
            if (playpos < write_ofs_)
            {
                /* 
                 *   the play cursor has wrapped - lock from the write
                 *   position to the end of the buffer, and from the start
                 *   of the buffer to the play position 
                 */
                locklen = (bufsize_ - write_ofs_) + playpos;
            }
            else
            {
                /* 
                 *   no wrapping has occurred - lock from the write
                 *   position to the play position 
                 */
                locklen = playpos - write_ofs_;
            }

            /* lock the block */
            if (istreambuf_->Lock(write_ofs_, locklen,
                                  (void **)&write1, &len1,
                                  (void **)&write2, &len2, 0) != DS_OK)
            {
                /* failure locking the block - abort */
                err = 1;
                break;
            }

            /* calculate the next read location */
            read_ofs_ += locklen;
            if (read_ofs_ >= bufsize_)
                read_ofs_ = 0;

            /* count the chunk we've finished */
            bytes_played_ += locklen;

            /*
             *   If we're done reading, check to see if the cursor has
             *   moved into the silent part yet.  If so, we're done
             *   playing back everything. 
             */
            if (done_reading_)
            {
                /* check to see if we're in the silence area yet */
                if (silence_start < silence_end
                    ? (playpos >= silence_start && playpos < silence_end)
                    : (playpos >= silence_start || playpos < silence_end))
                {
                    /*
                     *   We're done - we've played back everything and are
                     *   now playing in the silent filler after the end of
                     *   the file.  We can stop looping now. 
                     */
                    istreambuf_->Stop();
                    done = TRUE;
                    break;
                }

                /*
                 *   We're done reading the file, but we haven't finished
                 *   playing back the part we've already read.  Keep
                 *   going.  In case we don't manage to cancel playback
                 *   before moving off the end of the data we've read so
                 *   far (because our timeout interval makes us we wake up
                 *   after playback has already moved off the end -- this
                 *   possibility is the big reason we use the DirectX 5
                 *   notification scheme when DirectX 5 is available),
                 *   fill the writable area with silence. 
                 */
                FillMemory(write1, len1, silence);
                if (write2 != 0)
                    FillMemory(write2, len2, silence);

                /* 
                 *   advance the silent end offset to the end of the new
                 *   silent section 
                 */
                if (len2 != 0)
                    silence_end = len2;
                else
                    silence_end = write_ofs_ + len1;
            }
            else
            {
                /* read data into the first block */
                if (reader_->read_data(write1, len1, &bytes_read,
                                       repeat_, &repeats_done_))
                {
                    err = 3;
                    break;
                }
                
                /* note if we've reached the end of the data */
                if (bytes_read < len1)
                {
                    /* fill the remainder with silence */
                    FillMemory(write1 + len1, len1 - bytes_read, silence);

                    /* 
                     *   Since this was a partial read, note the bounds of
                     *   the silent stretch.  We'll use this to track when
                     *   playback enters this portion, so that we know
                     *   when we're done with playback.  
                     */
                    silence_start = write_ofs_ + bytes_read;
                    silence_end = write_ofs_ + len1;

                    /* note that we're done */
                    done_reading_ = TRUE;

                    /* 
                     *   if we have a second block, fill it with silence
                     *   as well 
                     */
                    if (write2)
                    {
                        /* fill the second block with silence */
                        FillMemory(write2, len2, silence);

                        /* advance the silence end pointer accordingly */
                        silence_end = len2;
                    }
                }
                else if (write2 != 0)
                {
                    /* 
                     *   We have a second block - fill it as well 
                     */
                    if (reader_->read_data(write2, len2, &bytes_read,
                                           repeat_, &repeats_done_))
                    {
                        err = 4;
                        break;
                    }

                    /* if we didn't fill the request, we're done */
                    if (bytes_read < len2)
                    {
                        /* fill the remainder with silence */
                        FillMemory(write2 + len2, len2 - bytes_read, silence);

                        /* 
                         *   Note the bounds of the silent stretch.  Note
                         *   that the second block always starts at the
                         *   beginning of the buffer. 
                         */
                        silence_start = 0;
                        silence_end = bytes_read;

                        /* note that we're done reading */
                        done_reading_ = TRUE;
                    }
                }
            }

            /* unlock the buffer */
            istreambuf_->Unlock(write1, len1, write2, len2);

            /* calculate the next location to write */
            write_ofs_ += len1 + len2;
            if (write_ofs_ >= bufsize_)
                write_ofs_ -= bufsize_;

            /* done */
            break;
        }
    }

    /* stop the buffer if there was an error */
    if (err != 0)
        istreambuf_->Stop();

    /* we can clean up the notification resources now */
    delete_notify_resources();

    /*
     *   If we have a window to notify on completion, post a message to it
     *   now telling it we're done.  
     */
    if (win_cb_ != 0)
        PostMessage(win_cb_, win_cb_msg_,
                    (WPARAM)repeats_done_, win_cb_lpar_);

    /* done - return the error code we set */
    return err;
}

/*
 *   Notification thread main entrypoint
 */
void CTadsWavFilePlayer::notify_thread_main(void *ctx)
{
    /* call the member function to handle the thread */
    ((CTadsWavFilePlayer *)ctx)->do_notify();
}

/*
 *   Notification thread main handler 
 */
DWORD CTadsWavFilePlayer::do_notify()
{
    DWORD wait_id;
    int done;
    int err;
    BYTE silence = (BYTE)(reader_->get_wavefmtex()
                          ->wBitsPerSample == 8 ? 128 : 0);


    /* keep servicing events until we finish playback */
    for (done = FALSE, err = 0 ; !done && err == 0 ; )
    {
        /* wait for the next event */
        wait_id = WaitForMultipleObjects(2, notify_evt_, FALSE, INFINITE);
        if (wait_id == WAIT_FAILED)
        {
            err = 1;
            break;
        }

        /* 
         *   Figure out which event occurred.  The return value from the
         *   wait call is the index into our array of event handles. 
         */
        switch(wait_id - WAIT_OBJECT_0)
        {
        case 0:
            /* 
             *   Chunk-end notification.  This event occurs every time
             *   playback reaches a chunk boundary, which indicates that
             *   it's time to fill up the chunk that was just finished
             *   with fresh data from the file.  
             */
            {
                BYTE *write1;
                DWORD len1;
                unsigned long bytes_read;

                /* calculate the next read location */
                read_ofs_ += chunksize_;
                if (read_ofs_ >= bufsize_)
                    read_ofs_ = 0;

                /* count the chunk we finished in the bytes played */
                bytes_played_ += chunksize_;

                /* check to see if we've finished reading the file */
                if (done_reading_ && read_ofs_ == last_block_ofs_)
                {
                    /* 
                     *   Tell the stream to stop playback.  Note that we
                     *   proceed to fill in the finished chunk with
                     *   silence; since sound playback is running
                     *   asynchronously, there's a chance that playback
                     *   will continue for a while after we request the
                     *   stop, so we may loop back into this chunk before
                     *   the hardware actually stops playing the buffer.  
                     */
                    istreambuf_->Stop();
                }
                
                /* lock the chunk that just finished */
                if (istreambuf_->Lock(write_ofs_, chunksize_,
                                      (void **)&write1, &len1,
                                      0, 0, 0) != DS_OK)
                {
                    err = 2;
                    break;
                }
                
                /* 
                 *   Load file data into the chunk.  If we've already
                 *   finished reading the file, there's nothing to read
                 *   here. 
                 */
                if (done_reading_)
                {
                    /* nothing left in file - we read zero bytes */
                    bytes_read = 0;

                    /* fill this block with silence */
                    FillMemory(write1, len1, silence);
                }
                else
                {
                    /* read from the file */
                    if (reader_->read_data(write1, len1, &bytes_read,
                                           repeat_, &repeats_done_))
                    {
                        err = 3;
                        break;
                    }

                    /*
                     *   If we didn't get as much as we asked for, note
                     *   the partial last block read length.  
                     */
                    if (bytes_read < len1)
                    {
                        /* 
                         *   Since this was a partial read, note the
                         *   partial read position.  When we reach this
                         *   playback position, we'll know we're done.  
                         */
                        last_block_ofs_ = write_ofs_ + chunksize_;
                        if (last_block_ofs_ >= bufsize_)
                            last_block_ofs_ = 0;

                        /* note that we've reached the end */
                        done_reading_ = TRUE;

                        /* 
                         *   fill the remainder of the last block with
                         *   silence 
                         */
                        FillMemory(write1 + bytes_read, len1 - bytes_read,
                                   silence);
                    }
                }
                
                /* unlock the buffer */
                istreambuf_->Unlock(write1, len1, 0, 0);
                
                /* calculate the next location to write */
                write_ofs_ += len1;
                if (write_ofs_ >= bufsize_)
                    write_ofs_ = 0;
            }
            break;

        case 1:
            /*
             *   Stop notification.  This tells us that playback has
             *   finished.  At this time, we can shut down the thread, so
             *   note that we're done.  
             */
            done = TRUE;
            break;
        }
    }

    /* if an error occurred, halt playback */
    if (err != 0)
        istreambuf_->Stop();
    
    /*
     *   We can now clean up the notification resources, since we're not
     *   going to need them any more. 
     */
    delete_notify_resources();

    /*
     *   If we have a window to notify on completion, post a message to it
     *   now telling it we're done. 
     */
    if (win_cb_ != 0)
        PostMessage(win_cb_, win_cb_msg_,
                    (WPARAM)repeats_done_, win_cb_lpar_);

    /* done - return the error code we set */
    return err;
}


/* ------------------------------------------------------------------------ */
/*
 *   Wave file reader implementation
 */

CTadsWavFileReader::CTadsWavFileReader()
{
    /* no header yet */
    wavefmt_ = 0;
}

CTadsWavFileReader::~CTadsWavFileReader()
{
    /* if we have allocated a header, free it now */
    if (wavefmt_ != 0)
        th_free(wavefmt_);
}

/*
 *   Read a WAV file header.  Before calling this, seek to the point in
 *   the file where the WAV data begins.  This parses the file header and
 *   sets up the WAVEFORMATEX structure.  Returns zero on success,
 *   non-zero on failure.  
 */
int CTadsWavFileReader::read_header(HANDLE hfile)
{
    unsigned char buf[12];
    unsigned long chunklen;
    unsigned long fpos, endpos;
    DWORD bytes_read;

    /* remember our file pointer */
    hfile_ = hfile;

    /* if we have a previous header allocated, free it */
    if (wavefmt_ != 0)
    {
        th_free(wavefmt_);
        wavefmt_ = 0;
    }
    
    /* read the RIFF header, first chunk length, and first chunk type */
    if (ReadFile(hfile_, buf, 12, &bytes_read, 0) == 0 || bytes_read != 12)
        return 1;

    /* 
     *   check the header - make sure it's a RIFF file whose first chunk
     *   is WAVE 
     */
    if (memcmp(buf, "RIFF", 4) != 0 || memcmp(buf + 8, "WAVE", 4) != 0)
        return 2;

    /* note our current location */
    fpos = SetFilePointer(hfile_, 0, 0, FILE_CURRENT);

    /* 
     *   Get the length of the WAVE chunk - it's a 4-byte little-endian
     *   value, just like our own native unsigned long type (as this is
     *   Windows-only code reading a Windows file format, what a
     *   coincidence!).  From the length, calculate the ending location.  
     */
    memcpy(&chunklen, buf + 4, 4);
    endpos = fpos + chunklen - 4;

    /* presume we won't find the subchunks */
    found_data_ = found_header_ = FALSE;

    /*
     *   Scan the WAVE chunk for subchunks.  It should have two: one
     *   labelled "fmt " with the WAV header, and one labelled "data" with
     *   the PCM byte stream.  We want to read the "fmt " header and
     *   remember the location of the "data" chunk so that we can stream
     *   it in for playback later.  
     */
    while (fpos < endpos)
    {
        unsigned long sublen;
            
        /* read the next chunk header (type code plus 4-byte length) */
        if (ReadFile(hfile_, buf, 8, &bytes_read, 0) == 0 || bytes_read != 8)
            return 3;

        /* get the subchunk length */
        memcpy(&sublen, buf + 4, 4);

        /* note the change in file position */
        fpos += 8;

        /* see what we have */
        if (memcmp(buf, "fmt ", 4) == 0)
        {
            PCMWAVEFORMAT pcmwavefmt;
            WORD extra_len;
            
            /* this is the header chunk */
            found_header_ = TRUE;

            /* 
             *   make sure it's big enough to be a WAV file - it has to at
             *   least contain a PCMWAVEFORMAT structure 
             */
            if (sublen < sizeof(PCMWAVEFORMAT))
                return 6;

            /* read the PCMWAVEFORMAT structure */
            if (ReadFile(hfile_, &pcmwavefmt, sizeof(pcmwavefmt),
                         &bytes_read, 0) == 0
                || bytes_read != sizeof(pcmwavefmt))
                return 7;

            /* 
             *   Check the format tag.  If it's WAVE_FORMAT_PCM, it means
             *   that the PCMWAVEFORMAT structure we just read is the
             *   whole thing.  Otherwise, read a 16-bit integer which will
             *   tell us how many extra bytes there are in the structure. 
             */
            if (pcmwavefmt.wf.wFormatTag == WAVE_FORMAT_PCM)
            {
                /* vanilla PCMWAVEFORMAT structure; nothing extra */
                extra_len = 0;
            }
            else
            {
                /* 
                 *   the next 16-bit word in the file is the number of
                 *   extra header bytes 
                 */
                if (ReadFile(hfile_, &extra_len, sizeof(extra_len),
                             &bytes_read, 0) == 0
                    || bytes_read != sizeof(extra_len))
                    return 8;
            }

            /* allocate space for the header structure */
            wavefmt_ = (WAVEFORMATEX *)
                       th_malloc(sizeof(WAVEFORMATEX) + extra_len);

            /* copy the PCMWAVEFORMAT base structure */
            memcpy(wavefmt_, &pcmwavefmt, sizeof(pcmwavefmt));

            /* read the extra bytes, if there are any */
            wavefmt_->cbSize = extra_len;
            if (extra_len != 0
                && (ReadFile(hfile_,
                             (((BYTE *)&wavefmt_->cbSize)
                              + sizeof(wavefmt_->cbSize)), extra_len,
                             &bytes_read, 0) == 0
                    || bytes_read != extra_len))
                return 9;
        }
        else if (memcmp(buf, "data", 4) == 0)
        {
            /* this is the data chunk - note its location */
            found_data_ = TRUE;
            data_fpos_ = fpos;
            data_len_ = sublen;
        }
        else
        {
            /* ignore other subchunk types */
        }

        /* seek to the start of the next subchunk */
        fpos += ((sublen + 1) & ~1);
        SetFilePointer(hfile_, fpos, 0, FILE_BEGIN);
    }

    /* if we didn't find either chunk, it's not a valid file */
    if (!found_data_ || !found_header_)
        return 5;

    /* success */
    return 0;
}

/*
 *   Read from the data section 
 */
int CTadsWavFileReader::read_data(unsigned char *buf,
                                  unsigned long bytes_to_read,
                                  unsigned long *bytes_read,
                                  int repeat, int *repeats_done)
{
    DWORD actual_read;
    unsigned long org_bytes_to_read = bytes_to_read;
    unsigned long avail;
    unsigned char *org_buf = buf;
    unsigned char *cur_buf;

    /* seek to the next read position */
    if (SetFilePointer(hfile_, data_fpos_ + data_read_ofs_, 0,
                       FILE_BEGIN) == 0xffffffff)
        return 1;

    /*
     *   Figure out how many bytes are available to be read.  We can read
     *   the amount remaining on this iteration, plus the full data size
     *   times the number of additional iterations remaining.  If we're
     *   repeating indefinitely, we can read any amount. 
     */
    if (repeat == 0)
    {
        /* repeating forever - we can read as much as they want */
        avail = bytes_to_read;
    }
    else
    {
        /* start off with the amount left on the current iteration */
        avail = data_len_ - data_read_ofs_;

        /* add additional iterations, if any are available */
        if (repeat > *repeats_done)
            avail += data_len_ * (repeat - *repeats_done - 1);
    }
    
    /* 
     *   Determine if we can satisfy the request, based on how many bytes
     *   we have remaining in the chunk.  If we can't satisfy the request,
     *   reduce the number of bytes to read to the number actually
     *   remaining, and load the bytes into the end of the buffer.  We
     *   load into the end of the buffer because of the way direct sound
     *   notifications work: we can only set up a notification at certain
     *   fixed points, so we need to align the end of the file against one
     *   of these fixed points.  The buffer that was passed in ends at one
     *   of these notification points, so put our data at the end of the
     *   caller's buffer.  
     */
    if (bytes_to_read > avail)
    {
        /* shorten the amount to read by the actual amount remaining */
        bytes_to_read = avail;
    }

    /* 
     *   Read the file.  First read as much as we have left on the current
     *   iteration, then loop back to the start of the file and read from
     *   the beginning until we satisfy the request, looping back again
     *   and again if necessary. 
     */
    for (cur_buf = buf, actual_read = 0 ; bytes_to_read != 0 ; )
    {
        unsigned long cur_len;
        unsigned long cur_actual;

        /* read up to the amount left on the current iteration */
        cur_len = bytes_to_read;
        if (cur_len > data_len_ - data_read_ofs_)
            cur_len = data_len_ - data_read_ofs_;

        /* read this block */
        if (ReadFile(hfile_, cur_buf, cur_len, &cur_actual, 0) == 0
            || cur_actual != cur_len)
            return 1;

        /* add this read into the total */
        actual_read += cur_actual;

        /* advance the read offset */
        data_read_ofs_ += cur_actual;

        /* advance the buffer pointer */
        cur_buf += cur_len;

        /* decrement the amount left to read */
        bytes_to_read -= cur_len;

        /* if we've exhausted the file, count the completed iteration */
        if (data_read_ofs_ == data_len_)
            ++(*repeats_done);

        /*
         *   if there's anything left to read, wrap to the start of the
         *   data -- the only way we will need to read in multiple blocks
         *   in this loop is if we need to loop back to the start of the
         *   file, so do so here
         */
        if (bytes_to_read != 0)
        {
            /* reset to the start of the data stream */
            data_read_ofs_ = 0;
            if (SetFilePointer(hfile_, data_fpos_, 0, FILE_BEGIN)
                == 0xffffffff)
                return 1;
        }
    }

    /* set the amount we read for the caller's information */
    *bytes_read = actual_read;

    /* success */
    return 0;
}


#if 0
/* ------------------------------------------------------------------------ */
/*
 *   Test section 
 */

static void errexit(const char *msg)
{
    printf("error: %s\n", msg);
    exit(1);
}

void main()
{
    osfildef *fp;
    CTadsWavFileReader *reader;
    int err;

    /* open a file */
    fp = osfoprb("c:\\win95\\msremind.wav", OSFTBIN);
    if (fp == 0)
        errexit("can't open file");

    /* create a reader */
    reader = new CTadsWavFileReader();

    /* read the header */
    if ((err = reader->read_header(fp)) != 0)
    {
        char buf[128];
        sprintf(buf, "return code %d reading header", err);
        errexit(buf);
    }

    /* done */
    delete reader;
    osfcls(fp);
}

void oshtml_dbg_printf(const char *, ...)
{
}


#endif
