/* $Header: d:/cvsroot/tads/html/win32/tadswav.h,v 1.1.1.1 1999/05/09 21:32:31 MJRoberts Exp $ */

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  tadswav.h - TADS WAV (waveform audio) support
Function
  
Notes
  
Modified
  01/19/98 MJRoberts  - Creation
*/

#ifndef TADSWAV_H
#define TADSWAV_H

#include <windows.h>
#include <dsound.h>

#ifndef TADSHTML_H
#include "tadshtml.h"
#endif


/*
 *   Wave file player.  This object allows playback of a WAV file, given a
 *   pointer to a file containing WAV-format data. 
 */
class CTadsWavFilePlayer
{
public:
    CTadsWavFilePlayer();
    ~CTadsWavFilePlayer();

    /* 
     *   Begin playing a WAV file.  Before calling, the client must seek
     *   the file to the start of the WAV header (for a .WAV file, this is
     *   simply the first byte in the file).
     *   
     *   Returns immediately after starting playback (playback continues
     *   in the background).  Returns zero on success, non-zero on failure.
     *   
     *   When playback is finished, we will post the given message, with
     *   the given LPARAM (and a WPARAM of zero), to the given window
     *   handle.  We use a message-based notifier, rather than a function
     *   callback, because wave file playback is handled by a separate
     *   thread; by using a message, we can avoid any reentrancy issues
     *   that might arise if we were to invoke a callback function
     *   directly from a background thread.  
     */
    int play(struct IDirectSound *ids, HANDLE hfile,
             HWND win_cb, UINT win_cb_msg, LPARAM win_cb_lpar, int repeat);

    /*
     *   Prepare for playback.  This allocates all resources and queues up
     *   the WAV file at the start of the data.  Returns zero on success,
     *   non-zero on failure.  If we're already prepared, we'll return
     *   success.  
     */
    int prepare(IDirectSound *ids, HANDLE hfile, HWND win_cb,
                UINT win_cb_msg, LPARAM win_cb_lpar, int repeat);

    /*
     *   Stop playback.  If 'sync' is true, it means that we must not
     *   return until the sound has actually stopped and the callback has
     *   been invoked; otherwise, we may return before the callback is
     *   invoked.  
     */
    void stop(int sync);

private:
    /* 
     *   load the initial bufferfull of data; returns zero on success,
     *   nonzero on failure 
     */
    int load_init_buffer();

    /*
     *   initialize notifications in preparation for playback
     */
    int init_notify();

    /* clean up old state, deleting any objects we've allocated */
    void clean_up();

    /* clean up resources used by the notification thread */
    void delete_notify_resources();

    /* notification thread main entrypoint */
    static void notify_thread_main(void *ctx);

    /* timer thread main entrypoint */
    static void timer_thread_main(void *ctx);
    

    /* 
     *   notification thread main routine (member version; the static
     *   version just translates its parameter into a CTadsWavFileReader
     *   and calls us) 
     */
    DWORD do_notify();

    /*
     *   timer thread main routine (member version; the static version
     *   calls this) 
     */
    DWORD do_timer();

    /* flag indicating that we've prepared resources for playback */
    int prepared_ : 1;
    
    /* our WAV file reader */
    class CTadsWavFileReader *reader_;

    /* size of the buffer */
    DWORD bufsize_;

    /* 
     *   number of iterations remaining to play back; 0 means that we keep
     *   looping indefinitely
     */
    int repeat_;

    /* number of iterations that we've finished so far */
    int repeats_done_;

    /* 
     *   size of the notification chunk - DirectSound will notify us when
     *   playback reaches the end of each chunk 
     */
    DWORD chunksize_;

    /* our stream buffer */
    struct IDirectSoundBuffer *istreambuf_;

    /* our notifier interface */
    struct IDirectSoundNotify *inotify_;

    /* 
     *   our DirectSound main interface - we need to hold onto this until
     *   all of our other direct sound interfaces are released, since the
     *   other interfaces seem to require it to exist at deletion time but
     *   don't set a reference on it themselves 
     */
    struct IDirectSound *idirsnd_;

    /* offset in buffer of next position to write/read */
    unsigned long write_ofs_;
    unsigned long read_ofs_;

    /* 
     *   flag indicating we've finished copying the file into the buffer
     *   -- this will occur before we've actually finished playing back
     *   the file, because playback lags loading (for obvious reasons) 
     */
    int done_reading_ : 1;

    /* flag indicating that we've finished playing back the entire file */
    int done_playing_ : 1;

    /*
     *   Size and position of the last block we read.  The last block may
     *   not completely fill the buffer, since the file may not be an even
     *   multiple of the buffer size.  When the last block doesn't
     *   completely fill the buffer, we'll read the data into the end of
     *   the buffer, so that it aligns against a notification point, and
     *   then we'll advance the read pointer when we get to the last
     *   buffer so that it points to the starting point.  This size
     *   records partial read length so that we can figure out where to
     *   put the read pointer when we get there (we load in advance of
     *   playing, so we need to keep track of this from the time we do the
     *   last buffer load until the time the playback catches up to that
     *   point).  
     */
    unsigned long last_block_ofs_;

    /* event handles - one for chunk notifications, one for stop */
    HANDLE notify_evt_[2];

    /* handle of notification thread */
    HANDLE thread_hdl_;

    /* ID of notification thread */
    DWORD thread_id_;

    /* 
     *   Number of bytes we've *finished* playing back so far.  Each time
     *   we are notified that a chunk has completed playing, we update
     *   this value.  
     */
    unsigned long bytes_played_;

    /* callback information */
    HWND win_cb_;
    UINT win_cb_msg_;
    LPARAM win_cb_lpar_;

    /* critical section synchronizer */
    CRITICAL_SECTION sync_;
};

/*
 *   Number of chunks into which to divide the playback buffer.
 *   DirectSound will notify us whenever playback reaches the end of one
 *   of the chunks, so that we can refill the chunk.  Use three chunks of
 *   one second each; this should give us enough of a cushion that we can
 *   keep up with refilling the buffers before they're needed, thus
 *   avoiding playback glitches.  
 */
const int TADSWAV_NOTIFY_CNT = 3;

/*
 *   Wave file reader.  This object parses a WAV file. 
 */
class CTadsWavFileReader
{
public:
    CTadsWavFileReader();
    ~CTadsWavFileReader();

    /*
     *   Read a WAV file header.  This parses the file header and sets up
     *   the WAVEFORMATEX structure.  Returns zero on success, non-zero on
     *   failure.  
     */
    int read_header(HANDLE hfile);

    /* 
     *   get a pointer to the WAVEFORMATEX structure (the contents of this
     *   structure aren't valid until the header has been read in) 
     */
    WAVEFORMATEX *get_wavefmtex() const { return wavefmt_; }

    /* seek to the start of the wave data stream */
    void seek_data_start() { data_read_ofs_ = 0; }

    /*
     *   Read from the wave data stream.  Returns zero on success, nonzero
     *   on error.  *bytes_read returns with the actual number of bytes
     *   that we read; if we reach the end of the file before the full
     *   request is satisfied, we'll return success with *bytes_read
     *   indicating that the buffer is only partially filled. 
     */
    int read_data(unsigned char *buf, unsigned long bytes_to_read,
                  unsigned long *bytes_read, int repeat, int *repeats_done);

    /* get the total length of the wave data byte stream */
    unsigned long get_wave_len() const { return data_len_; }

private:
    /* our file pointer */
    HANDLE hfile_;
    
    /*
     *   Windows wave file descriptor structure.  This contains the header
     *   information from the wave file.  
     */
    WAVEFORMATEX *wavefmt_;

    /* 
     *   seek position in file of start of byte stream containing the
     *   actual digitized sound data 
     */
    unsigned long data_fpos_;

    /* flags: we've found the header and data chunks in the file */
    int found_header_ : 1;
    int found_data_ : 1;

    /*
     *   Data read offset - this is the byte offset in the data stream
     *   from which the next data read routine will start reading. 
     */
    unsigned long data_read_ofs_;

    /* size of the data chunk */
    unsigned long data_len_;
};

#endif /* TADSWAV_H */


